import "cross-fetch";
import "cross-fetch/polyfill";
import * as graph from "@microsoft/microsoft-graph-client";
import * as id from "@azure/identity";
import * as common from "@data-heaving/common";
import * as fs from "fs/promises";
import * as ad from "./run-ad";
import * as auth from "./run-auth";
import * as pulumiSetup from "./run-pulumi";
import * as certs from "./run-certs";
export const performBootstrap = async (inputs) => {
    const bootstrapperInfo = await setupBootstrapperApp(inputs);
    const { backendConfig, backendStorageAccountKey, cicdRGName, kvName } = await runWithBootstrapper(inputs, bootstrapperInfo);
    const { bootstrapperPulumiAuth } = bootstrapperInfo;
    if (bootstrapperPulumiAuth.type === "sp") {
        bootstrapperPulumiAuth.backendStorageAccountKey = backendStorageAccountKey;
    }
    return {
        cicdRGName,
        kvName,
        pulumiConfigInfo: {
            backendConfig,
            auth: bootstrapperPulumiAuth,
        },
        envSpecificPipelineConfigReader: bootstrapperInfo.envSpecificPipelineConfigReader,
    };
};
const setupBootstrapperApp = async ({ eventEmitter, credentials: { credentials, givenClientId }, bootstrapperApp, azure: { tenantId, subscriptionId }, }) => {
    // 1. Ensure bootstrap SP is created using current Azure CLI credentials
    let clientId;
    let principalId;
    let bootstrapperCredentials;
    let bootstrapperPulumiAuth;
    let spAuthStorageConfig;
    const graphClient = graph.Client.initWithMiddleware({
        authProvider: credentials,
        debugLogging: true, // This will print URL of every command
    });
    const envSpecificPipelineConfigReader = bootstrapperApp.type === "msi"
        ? {
            principalId: bootstrapperApp.principalId,
            principalType: "ServicePrincipal", // It appears that in Azure role assignments, even MSIs will get this principal type instead of "MSI",
        }
        : await ad.getCurrentPrincipalInfo(graphClient, givenClientId);
    eventEmitter.emit("afterResolvingConfigReaderPrincipal", common.deepCopy(envSpecificPipelineConfigReader));
    switch (bootstrapperApp.type) {
        case "sp":
            {
                const { keyPath, certPath, rsaBits, pfxPath, pfxPassword, ...spAuth } = bootstrapperApp.authentication;
                await certs.ensureKeyAndCertExists(keyPath, certPath, rsaBits, spAuth.certValidityPeriodDays, spAuth.certSubject);
                const certPEM = certs.ensureEndsWithNewline(await fs.readFile(certPath, "utf-8"));
                const clientAndPrincipalID = await ad.ensureBootstrapSPIsConfigured({
                    eventEmitter,
                    graphClient,
                    bootstrapperApp,
                    certificatePEM: certPEM,
                });
                ({ clientId, principalId } = clientAndPrincipalID);
                const { keyPEM, keyAndCertPath } = await certs.ensureCertificateCredentialsFileExists(spAuth.tempDir, keyPath, certPEM);
                spAuthStorageConfig = {
                    keyPEM,
                    certPEM,
                    configReaderPrincipalId: envSpecificPipelineConfigReader.principalId,
                    configSecretName: bootstrapperApp.configSecretName,
                };
                bootstrapperCredentials = new id.ClientCertificateCredential(tenantId, clientId, keyAndCertPath);
                await certs.ensurePfxExists(keyPath, certPath, pfxPath, pfxPassword);
                bootstrapperPulumiAuth = {
                    type: "sp",
                    clientId,
                    pfxPath,
                    pfxPassword,
                    backendStorageAccountKey: "", // Will be modified by performSetUp
                };
            }
            break;
        case "msi":
            {
                ({ clientId, principalId } = bootstrapperApp);
                bootstrapperCredentials = new id.ManagedIdentityCredential(clientId);
                bootstrapperPulumiAuth = {
                    type: "msi",
                    clientId,
                };
            }
            break;
        default:
            throw new Error(`Unrecognized bootstrapper app kind: "${bootstrapperApp.type}"`);
    }
    // 2. Ensure bootstrap SP has enough permissions to continue operating
    await auth.ensureBootstrapSPHasEnoughPrivileges({
        eventEmitter,
        credentials,
        principalId,
        subscriptionId,
    });
    return {
        bootstrapperCredentials,
        bootstrapperPulumiAuth,
        principalId,
        spAuthStorageConfig,
        envSpecificPipelineConfigReader,
    };
};
const runWithBootstrapper = async ({ eventEmitter, azure, organization, pulumiEncryptionKeyBits }, { bootstrapperCredentials, principalId, spAuthStorageConfig, }) => {
    // 4. Create credentials for bootstrap SP, and then create necessary resources for Pulumi state management using bootstrap SP
    return await pulumiSetup.ensureRequireCloudResourcesForPulumiStateExist({
        eventEmitter,
        credentials: bootstrapperCredentials,
        azure,
        organization,
        principalId,
        spAuthStorageConfig,
        pulumiEncryptionKeyBits,
    });
};
export const constructVaultName = pulumiSetup.constructVaultName;
export const tryGetSecretValue = pulumiSetup.tryGetSecretValue;
export const BEGIN_CERTIFICATE = certs.BEGIN_CERTIFICATE;
//# sourceMappingURL=run.js.map